@@ -191,6 +191,29 @@ void Buffer::Empty()
       save=true;
 }
 
+// move data from other buffer, prepare for SpaceAdd.
+int Buffer::MoveDataHere(Buffer *o,int max_len)
+{
+   const char *b;
+   int size;
+   o->Get(&b,&size);
+   if(size>max_len)
+      size=max_len;
+   if(size>0) {
+      if(size>=64 && Size()==0 && o->Size()==size && !save && !o->save) {
+	 // optimization by swapping buffers
+	 buffer.swap(o->buffer);
+	 buffer_ptr=replace_value(o->buffer_ptr,buffer_ptr);
+	 buffer.set_length_no_z(buffer_ptr);
+	 o->pos+=size;
+      } else {
+	 memcpy(GetSpace(size),b,size);
+	 o->Skip(size);
+      }
+   }
+   return size;
+}
+
 Buffer::Buffer()
 {
    saved_errno=0;
@@ -243,6 +266,20 @@ void DataTranslator::AppendTranslated(Buffer *target,const char *put_buf,int siz
    target->SetPos(old_pos);
 }
 
+void DirectedBuffer::SetTranslator(DataTranslator *t)
+{
+   if(mode==GET && !translator && Size()>0) {
+      // translate unread data
+      const char *data;
+      int len;
+      Get(&data,&len);
+      t->Put(data,len);
+      buffer.truncate(buffer_ptr);
+      t->AppendTranslated(this,0,0);
+   }
+   translator=t;
+}
+
 #ifdef HAVE_ICONV
 void DataRecoder::PutTranslated(Buffer *target,const char *put_buf,int size)
 {
@@ -319,34 +356,24 @@ DataRecoder::DataRecoder(const char *from_code,const char *to_code,bool translit
    if(translit) {
       const char *to_code_translit=xstring::cat(to_code,"//TRANSLIT",NULL);
       backend_translate=iconv_open(to_code_translit,from_code);
-      if(backend_translate!=(iconv_t)-1)
+      if(backend_translate!=(iconv_t)-1) {
+	 Log::global->Format(9,"initialized translation from %s to %s\n",from_code,to_code_translit);
 	 return;
+      }
       backend_translate=0;
    }
 
    backend_translate=iconv_open(to_code,from_code);
-   if(backend_translate!=(iconv_t)-1)
+   if(backend_translate!=(iconv_t)-1) {
+      Log::global->Format(9,"initialized translation from %s to %s\n",from_code,to_code);
       return;
+   }
 
    Log::global->Format(0,"iconv_open(%s,%s) failed: %s\n",
 			      to_code,from_code,strerror(errno));
    backend_translate=0;
 }
 
-void DirectedBuffer::SetTranslator(DataTranslator *t)
-{
-   if(mode==GET && !translator && Size()>0) {
-      // translate unread data
-      const char *data;
-      int len;
-      Get(&data,&len);
-      t->Put(data,len);
-      buffer.truncate(buffer_ptr);
-      t->AppendTranslated(this,0,0);
-   }
-   translator=t;
-}
-
 void DirectedBuffer::SetTranslation(const char *enc,bool translit)
 {
    if(!enc || !*enc)
@@ -356,6 +383,8 @@ void DirectedBuffer::SetTranslation(const char *enc,bool translit)
       return;
    const char *from_code=(mode==GET?enc:local_code);
    const char *to_code  =(mode==GET?local_code:enc);
+   if(!strcasecmp(from_code,to_code))
+      return;
    SetTranslator(new DataRecoder(from_code,to_code,translit));
 }
 #endif //HAVE_ICONV
@@ -379,6 +408,16 @@ void DirectedBuffer::PutTranslated(const char *buf,int size)
    else
       Buffer::Put(buf,size);
 }
+int DirectedBuffer::MoveDataHere(Buffer *buf,int size)
+{
+   if(size>buf->Size())
+      size=buf->Size();
+   if(mode==PUT && translator)
+      translator->PutTranslated(this,buf->Get(),size);
+   else
+      return Buffer::MoveDataHere(buf,size);
+   return size;
+}
 
 void DirectedBuffer::EmbraceNewData(int len)
 {
@@ -398,7 +437,8 @@ void DirectedBuffer::EmbraceNewData(int len)
 
 
 IOBuffer::IOBuffer(dir_t m)
-   : DirectedBuffer(m), event_time(now), max_buf(0)
+   : DirectedBuffer(m), event_time(now),
+     max_buf(0), get_size(GET_BUFSIZE)
 {
 }
 IOBuffer::~IOBuffer()
@@ -428,6 +468,18 @@ void IOBuffer::Put(const char *buf)
    Put(buf,strlen(buf));
 }
 
+int IOBuffer::TuneGetSize(int res)
+{
+   if(res>0)
+   {
+      // buffer size tuning depending on data rate
+      const int max_get_size=(max_buf?max_buf:0x100000);
+      if(res>get_size/2 && Size()+get_size*2<=max_get_size)
+	 get_size*=2;
+   }
+   return res;
+}
+
 int IOBuffer::Do()
 {
    if(Done() || Error())
@@ -451,7 +503,7 @@ int IOBuffer::Do()
    case GET:
       if(eof)
 	 return STALL;
-      res=Get_LL(GET_BUFSIZE);
+      res=TuneGetSize(Get_LL(get_size));
       if(res>0)
       {
 	 EmbraceNewData(res);
@@ -502,15 +554,13 @@ int IOBufferStacked::Do()
       {
 	 buffer_ptr+=res;
 	 m=MOVED;
-	 down->Do();
       }
       break;
 
    case GET:
-      m|=down->Do();
       if(eof)
 	 return m;
-      res=Get_LL(GET_BUFSIZE);
+      res=Get_LL(/*unused*/0);
       if(res>0)
       {
 	 EmbraceNewData(res);
@@ -547,19 +597,9 @@ int IOBufferStacked::Get_LL(int)
       return 0;
    }
    down->ResumeSlave();
-
-   const char *b;
-   int size;
-   down->Get(&b,&size);
-
-   if(!b)
-   {
-      eof=true;
-      return 0;
-   }
-
-   memcpy(GetSpace(size),b,size);
-   down->Skip(size);
+   int size=MoveDataHere(down,down->Size());
+   if(down->Size()==0 && down->Eof())
+      PutEOF();
    return size;
 }
 
@@ -654,12 +694,19 @@ int IOBufferFDStream::Get_LL(int size)
       return 0;
    }
 
+   if(!Ready(fd,POLLIN))
+   {
+      Block(fd,POLLIN);
+      return 0;
+   }
+
    res=read(fd,GetSpace(size),size);
    if(res==-1)
    {
       saved_errno=errno;
       if(E_RETRY(saved_errno))
       {
+	 SetNotReady(fd,POLLIN);
 	 Block(fd,POLLIN);
 	 return 0;
       }
@@ -669,8 +716,10 @@ int IOBufferFDStream::Get_LL(int size)
       goto stream_err;
    }
 
-   if(res==0)
+   if(res==0) {
+      Log::global->Format(10,"buffer: EOF on FD %d\n",fd);
       eof=true;
+   }
    return res;
 
 stream_err:
@@ -710,7 +759,7 @@ int IOBufferFileAccess::Get_LL(int size)
 
    int res=0;
 
-   res=session->Read(GetSpace(size),size);
+   res=session->Read(this,size);
    if(res<0)
    {
       if(res==FA::DO_AGAIN)
